mod daemon;
mod listenbrainz;
mod models;

use navidrome::{client::builder::NavidromeBuilder, models::navidrome::NavidromeTrack};

use crate::daemon::{update_playlists_daemon, update_tracks_daemon};
use crate::models::{PartialTrackedPlaylist, TrackedPlaylist};
use axum::extract::State;
use axum::routing::{get, put};
use axum::{extract, Json, Router};
use serde::Serialize;
use sqlx::postgres::PgPool;
use sqlx::Error;
use std::collections::HashMap;
use std::env;
use std::net::SocketAddr;
use std::sync::Arc;
use tokio::sync::RwLock;

#[derive(Clone)]
struct TrackState {
    tracks: Arc<RwLock<HashMap<String, NavidromeTrack>>>,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let database = PgPool::connect(&env::var("DATABASE_URL")?).await?;
    sqlx::migrate!().run(&database).await?;

    let media_client = NavidromeBuilder::new()
        .base(env::var("NAVIDROME_BASE")?)
        .token(env::var("NAVIDROME_TOKEN")?)
        .build()?;

    let mut tracks = Arc::new(RwLock::new(HashMap::new()));

    let state = TrackState {
        tracks: tracks.clone(),
    };

    tokio::spawn(async move {
        update_playlists_daemon().await.unwrap();
    });

    tokio::spawn(async move {
        update_tracks_daemon(&media_client, &mut tracks)
            .await
            .unwrap();
    });

    let app = Router::new()
        .route("/playlists", get(all_playlists))
        .route("/playlists", put(create_playlist))
        .with_state(database)
        .with_state(state);

    let addr = SocketAddr::from(([127, 0, 0, 1], 3000));
    axum::Server::bind(&addr)
        .serve(app.into_make_service())
        .await?;

    Ok(())
}

#[derive(Serialize)]
struct JsonError {
    error: String,
}

impl From<Error> for JsonError {
    fn from(value: Error) -> Self {
        Self {
            error: value.to_string(),
        }
    }
}

#[derive(Serialize)]
enum Response<T> {
    Success(T),
    Error(JsonError),
}

async fn create_playlist(
    State(pool): State<PgPool>,
    extract::Json(partial): extract::Json<PartialTrackedPlaylist>,
) -> Json<Response<TrackedPlaylist>> {
    match partial.record(&pool).await {
        Ok(playlist) => Json(Response::Success(playlist)),
        Err(e) => Json(Response::Error(e.into())),
    }
}

async fn all_playlists(State(pool): State<PgPool>) -> Json<Vec<TrackedPlaylist>> {
    match TrackedPlaylist::all(&pool).await {
        Ok(playlists) => Json(playlists),
        Err(_) => Json(vec![]),
    }
}
